<?php

namespace DHLParcel\Shipping\Model\Service\Logic;

use DHLParcel\Shipping\Model\Core\Kernel;
use DHLParcel\Shipping\Model\Api\Connector;
use DHLParcel\Shipping\Model\Core\SingletonAbstract;
use DHLParcel\Shipping\Model\UUID;
use DHLParcel\Shipping\Model\Data\Api\Request\Shipment as ShipmentRequest;
use DHLParcel\Shipping\Model\Data\Api\General\Addressee;
use DHLParcel\Shipping\Model\Data\Api\General\Addressee\Address;
use DHLParcel\Shipping\Model\Data\Api\Request\Shipment\Option;
use DHLParcel\Shipping\Model\Data\Api\Request\Shipment\Piece as PieceRequest;
use DHLParcel\Shipping\Model\Data\Api\Response\Shipment as ShipmentResponse;
use DHLParcel\Shipping\Model\Data\Api\Response\Capability\Option as CapabilityOption;
use OrderState;
use Order;
use Address as PrestaShopAddress;
use Country;
use DHLParcel\Shipping\Model\Core\Settings;
use Configuration;
use Customer;
use Context;

class ShipmentLogic extends SingletonAbstract
{
    /**
     * @param $orderId
     * @param array $options
     * @param array $pieces
     * @param bool $isBusiness
     * @return ShipmentRequest
     */
    public function getRequestData($orderId, $labelSizes, $options, $optionsData, $business, $weight)
    {
        $randomUUID = UUID::v4();
        $order = new Order($orderId);

        $receiverAddress = new PrestaShopAddress($order->id_address_delivery);
        $receiver = $this->getReceiverAddress($receiverAddress, $business);
        if (Kernel::instance()->isMultistoreActive() && Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_MULTISTORE_OVERRIDE, $order->id_shop)) {
            $shipper = $this->getShipperAddress($order->id_shop);
        } else {
            $shipper = $this->getShipperAddress();
        }

        $accountId = Configuration::get(Settings::API_ACCOUNT_ID);
        $options = $this->getRequestOptions($options, $optionsData);
        $pieces = [];
        foreach ($labelSizes as $labelSize) {
            if (empty($labelSize)) {
                continue;
            }
            $pieces = [];
            $piece = new PieceRequest();
            $piece->parcelType = $labelSize;
            $piece->quantity = 1;
            $piece->weight = $weight;
            $pieces[] = $piece;
        }

        /** @var ShipmentRequest $shipmentRequest */
        $shipmentRequest = new ShipmentRequest();
        $shipmentRequest->shipmentId = $randomUUID;
        $shipmentRequest->orderReference = (string)$orderId;
        $shipmentRequest->receiver = $receiver;
        $shipmentRequest->shipper = $shipper;
        $shipmentRequest->accountId = $accountId;
        $shipmentRequest->options = $options;
        $shipmentRequest->pieces = $pieces;
        $shipmentRequest->application = 'prestashop';

        $useHideShipper = count(array_filter($options, function ($option) {
            return $option->key === 'SSN';
        })) > 0;
        if ($useHideShipper) {
            $shipmentRequest->onBehalfOf = $shipmentRequest->shipper;
        }

        return $shipmentRequest;
    }

    /**
     * @param ShipmentRequest $shipmentRequest
     * @return ShipmentRequest
     */
    public function getReturnRequestData($shipmentRequest)
    {
        // Check for alternative return address with settings
//        if ($this->helper->getConfigData('shipper/alternative_return_address')) {
//            $receiver = $this->getShipperAddress('return');
//        } elseif (!empty($shipmentRequest->onBehalfOf)) {
//            $receiver = $shipmentRequest->onBehalfOf;
//        } else {
            $receiver = $shipmentRequest->shipper;
//        }
//
//        // Check if there is an 'onBehalfOf' and unset it
//        if (!empty($shipmentRequest->onBehalfOf)) {
//            $shipmentRequest->onBehalfOf = null;
//        }
//
        $shipper = $shipmentRequest->receiver;
        $randomUUID = UUID::v4();
//
        $shipmentRequest->shipmentId = $randomUUID;
        $shipmentRequest->receiver = $receiver;
        $shipmentRequest->shipper = $shipper;
        $shipmentRequest->returnLabel = true;
//
        // Return labels are DOOR only with no other options
        $shipmentRequest->options = [
            new Option(['key' => CapabilityOption::KEY_DOOR]),
        ];

        return $shipmentRequest;
    }

    /**
     * @param ShipmentRequest $shipmentRequest
     * @return ShipmentResponse|null
     */
    public function sendRequest($shipmentRequest)
    {
        $response = Connector::instance()->post('shipments', $shipmentRequest->toArray(true));
        if (!$response) {
            return null;
        }

        /** @var ShipmentResponse $shipmentResponse */
        $shipmentResponse = new ShipmentResponse($response);
        $shipmentResponse = $this->updatePieces($shipmentResponse, $shipmentRequest);

        // Enrich pieces with postalCode
        $shipmentResponse = $this->tagPostalCode($shipmentResponse, $shipmentRequest->receiver->address->postalCode);
        return $shipmentResponse;
    }

    /**
     * @param ShipmentResponse $shipmentResponse
     * @param ShipmentRequest $shipmentRequest
     * @return ShipmentResponse
     */
    protected function updatePieces($shipmentResponse, $shipmentRequest)
    {
        $options = [];
        foreach ($shipmentRequest->options as $option) {
            $options[] = $option->key;
        }
        $options = array_unique($options);

        if ($shipmentRequest->returnLabel) {
            $options = [CapabilityOption::KEY_ADD_RETURN_LABEL];
        }

        $pieces = [];
        foreach ($shipmentResponse->pieces as $piece) {
            $piece->options = $options;
            $pieces[] = $piece;
        }
        $shipmentResponse->pieces = $pieces;

        return $shipmentResponse;
    }

    protected function tagPostalCode(ShipmentResponse $shipmentResponse, $postalCode)
    {
        if (!empty($shipmentResponse) && !empty($shipmentResponse->pieces)) {
            $updatedPieces = [];
            foreach ($shipmentResponse->pieces as $piece) {
                $piece->postalCode = strtoupper(trim($postalCode));
                $updatedPieces[] = $piece;
            }
            $shipmentResponse->pieces = $updatedPieces;
        }
        return $shipmentResponse;
    }

    /**
     * @param string $group
     * @return Addressee
     */
    protected function getShipperAddress($shopId = null)
    {
        /** @var Addressee $addressee */
        $addressee = new Addressee([
            'name'        => [
                'firstName'   => Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_FIRST_NAME, $shopId),
                'lastName'    => Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_LAST_NAME, $shopId),
                'companyName' => Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_COMPANY, $shopId),
            ],
            'address'     => [
                'countryCode' => Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_COUNTRY, $shopId),
                'postalCode'  => strtoupper(Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_POSTCODE, $shopId)),
                'city'        => Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_CITY, $shopId),
                'street'      => Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_STREET, $shopId),
                'number'      => Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_STREET_NUMBER, $shopId),
                'isBusiness'  => true,
                'addition'    => Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_STREET_ADDITION, $shopId),
            ],
            'email'       => Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_EMAIL, $shopId),
            'phoneNumber' => Kernel::instance()->getMultistoreConfiguration(Settings::SHIPPING_ADDRESS_PHONE, $shopId),
        ]);

        return $addressee;
    }

    /**
     * @param PrestaShopAddress
     * @param bool $isBusiness
     * @return Addressee
     */
    protected function getReceiverAddress($address, $business)
    {
        $customer = new Customer($address->id_customer);

        /** @var Addressee $addressee */
        $addressee = new Addressee([
            'name'        => [
                'firstName'   => $address->firstname,
                'lastName'    => $address->lastname,
                'companyName' => $address->company ?: null,
            ],
            'address'     => [
                'countryCode' => Country::getIsoById($address->id_country),
                'postalCode'  => strtoupper($address->postcode),
                'city'        => $address->city,
                'street'      => '',
                'number'      => '',
                'isBusiness'  => $business,
                'addition'    => '',
            ],
            'email'       => $customer->email,
            'phoneNumber' => $address->phone,
        ]);

        $addressee->address = $this->updateAddressStreet($addressee->address, [$address->address1, $address->address2]);
        return $addressee;
    }

    /**
     * @param Address $address
     * @param array $street
     * @return Address
     */
    protected function updateAddressStreet(Address $address, array $street)
    {
        $fullStreet = implode(' ', $street);

        $data = $this->parseStreetData($fullStreet);
        $address->street = $data['street'];
        $address->number = $data['number'];
        $address->addition = $data['addition'];

        return $address;
    }

    /**
     * @param $raw
     * @return array [
     *      'street'   => (string) Parsed street $raw
     *      'number'   => (string) Parsed number from $raw
     *      'addition' => (string) Parsed additional street data from $raw
     * ]
     */
    protected function parseStreetData($raw)
    {
        $skipAdditionCheck = false;
        $skipReverseCheck = false;
        $cutout = '';

        // Cutout starting special numbers from regular parsing logic
        $parsableParts = explode(' ', trim($raw), 2);

        // Check if it has a number with letters
        if (preg_match('/[0-9]+[a-z]+/i', reset($parsableParts)) === 1) {
            $cutout = reset($parsableParts) . ' ';
            $skipReverseCheck = true;
            unset($parsableParts[0]);

            // Check if it has a number with more than just letters, but also other available numbers
        } elseif (preg_match('/[0-9]+[^0-9]+/', reset($parsableParts)) === 1 && preg_match('/\d/', end($parsableParts)) === 1) {
            $cutout = reset($parsableParts) . ' ';
            $skipReverseCheck = true;
            unset($parsableParts[0]);

            // Check if it has something before a number
        } elseif (preg_match('/[^0-9]+[0-9]+/', reset($parsableParts)) === 1) {
            $cutout = reset($parsableParts) . ' ';
            $skipReverseCheck = true;
            unset($parsableParts[0]);

            // Check if starts with number (with anything), but also has numbers in the rest of the address
        } elseif (preg_match('/[^0-9]*[0-9]+[^0-9]*/', reset($parsableParts)) === 1 && preg_match('/\d/', end($parsableParts)) === 1) {
            $cutout = reset($parsableParts) . ' ';
            $skipReverseCheck = true;
            unset($parsableParts[0]);
        }

        $parsableStreet = implode(' ', $parsableParts);

        preg_match('/([^0-9]*)\s*(.*)/', trim($parsableStreet), $streetParts);
        $address = [
            'street' => isset($streetParts[1]) ? trim($streetParts[1]) : '',
            'number' => isset($streetParts[2]) ? trim($streetParts[2]) : '',
            'addition' => '',
        ];

        // Check if $street is empty
        if (strlen($address['street']) === 0 && !$skipReverseCheck) {
            // Try a reverse parse
            preg_match('/([\d]+[\w.-]*)\s*(.*)/i', trim($parsableStreet), $streetParts);
            $address['street'] = isset($streetParts[2]) ? trim($streetParts[2]) : '';
            $address['number'] = isset($streetParts[1]) ? trim($streetParts[1]) : '';
            $skipAdditionCheck = true;
        }

        // Check if $number has no numbers
        if (preg_match('/\d/', $address['number']) === 0) {
            $address['street'] = trim($parsableStreet);
            $address['number'] = '';

            // Addition check
        } elseif (!$skipAdditionCheck) {
            // If there are no letters, but has additional spaced numbers, use last number as number, no addition
            preg_match('/([^a-z]+)\s+([\d]+)$/i', $address['number'], $number_parts);
            if (isset($number_parts[2])) {
                $address['street'] .= ' ' . $number_parts[1];
                $address['number'] = $number_parts[2];

                // Regular number / addition split
            } else {
                preg_match('/([\d]+)[ .-]*(.*)/i', $address['number'], $number_parts);
                $address['number'] = isset($number_parts[1]) ? trim($number_parts[1]) : '';
                $address['addition'] = isset($number_parts[2]) ? trim($number_parts[2]) : '';
            }
        }

        // Reassemble street
        if (isset($address['street'])) {
            $address['street'] = $cutout . $address['street'];
        }

        // Be sure these fields are filled
        $address['number'] = isset($address['number']) ? $address['number'] : '';
        $address['addition'] = isset($address['addition']) ? $address['addition'] : '';

        // Clean any starting punctuations
        preg_match('/^[[:punct:]\s]+(.*)/', $address['street'], $cleanStreet);
        if (isset($cleanStreet[1])) {
            $address['street'] = $cleanStreet[1];
        }
        preg_match('/^[[:punct:]\s]+(.*)/', $address['number'], $cleanNumber);
        if (isset($cleanNumber[1])) {
            $address['number'] = $cleanNumber[1];
        }
        preg_match('/^[[:punct:]\s]+(.*)/', $address['addition'], $cleanAddition);
        if (isset($cleanAddition[1])) {
            $address['addition'] = $cleanAddition[1];
        }

        return $address;
    }

    public function updateShippedStatus($orderId)
    {
        if (!Configuration::get(Settings::AFTER_ORDER_STATE_UPDATE)) {
            return false;
        }

        $order = new \Order($orderId);
        if (!$order) {
            return false;
        }

        // $order->hasBeenShipped() does not work and is out of sync with what's actually in the database
        // Hence we copy the function and modify it correctly (fortunately everything is 'public'):
        if ($order->getCurrentState() == Configuration::get(Settings::AFTER_ORDER_STATE_UPDATE)) {
            return false;
        }

        $id_employee = Context::getContext()->employee->id;

        $order->setCurrentState(Configuration::get(Settings::AFTER_ORDER_STATE_UPDATE), $id_employee);
        return true;
    }

    public function getRequestOptions($keys = [], $data = [])
    {
        $requestOptions = [];
        foreach ($keys as $key) {
            $option = new Option([
                'key' => $key,
            ]);
            if (array_key_exists($key, $data)) {
                $option->input = $data[$key];
            }

            $requestOptions[] = $option;
        }

        return $requestOptions;
    }
}
